# Copyright 2015 Huawei Technologies India Pvt. Ltd.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sys

import mock

from neutronclient import shell
from neutronclient.tests.unit import test_cli20

from networking_sfc.cli import port_pair as pp

from oslo_utils import uuidutils

ingress_port_UUID = uuidutils.generate_uuid()
egress_port_UUID = uuidutils.generate_uuid()


class CLITestV20PortPairExtensionJSON(test_cli20.CLITestV20Base):
    def setUp(self):
        super(CLITestV20PortPairExtensionJSON, self).setUp()
        self._mock_extension_loading()
        self.register_non_admin_status_resource('port_pair')

    def _create_patch(self, name, func=None):
        patcher = mock.patch(name)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing

    def _mock_extension_loading(self):
        ext_pkg = 'neutronclient.common.extension'
        port_pair = self._create_patch(ext_pkg +
                                       '._discover_via_entry_points')
        port_pair.return_value = [("port_pair", pp)]
        return port_pair

    def test_ext_cmd_loaded(self):
        neutron_shell = shell.NeutronShell('2.0')
        ext_cmd = {'port-pair-list': pp.PortPairList,
                   'port-pair-create': pp.PortPairCreate,
                   'port-pair-update': pp.PortPairUpdate,
                   'port-pair-delete': pp.PortPairDelete,
                   'port-pair-show': pp.PortPairShow}
        for cmd_name, cmd_class in ext_cmd.items():
            found = neutron_shell.command_manager.find_command([cmd_name])
            self.assertEqual(cmd_class, found[0])

    def test_create_port_pair_with_mandatory_param(self):
        """Create port_pair: myname."""
        resource = 'port_pair'
        cmd = pp.PortPairCreate(test_cli20.MyApp(sys.stdout), None)
        name = 'myname'
        myid = 'myid'
        args = [name, '--ingress', ingress_port_UUID,
                '--egress', egress_port_UUID]
        position_names = ['name', 'ingress', 'egress']
        position_values = [name, ingress_port_UUID, egress_port_UUID]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_create_port_pair_with_bidirectional_port(self):
        """Create port_pair: myname with bidirectional port."""
        resource = 'port_pair'
        cmd = pp.PortPairCreate(test_cli20.MyApp(sys.stdout), None)
        name = 'myname'
        myid = 'myid'
        args = [name, '--ingress', ingress_port_UUID,
                '--egress', ingress_port_UUID]
        position_names = ['name', 'ingress', 'egress']
        position_values = [name, ingress_port_UUID, ingress_port_UUID]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_create_port_pair_with_all_param(self):
        """Create port_pair: myname with all parameter"""
        resource = 'port_pair'
        cmd = pp.PortPairCreate(test_cli20.MyApp(sys.stdout),
                                None)
        name = 'myname'
        myid = 'myid'
        desc = "my_port_pair"
        service_fn_param = 'correlation=None,weight=2'
        service_fn_param_exp = {"correlation": "None", "weight": "2"}
        args = [name, '--ingress', ingress_port_UUID,
                '--egress', egress_port_UUID, '--description', desc,
                '--service-function-parameters', service_fn_param]
        position_names = ['name', 'ingress', 'egress', 'description',
                          'service_function_parameters']
        position_values = [name, ingress_port_UUID, egress_port_UUID, desc,
                           service_fn_param_exp]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_update_port_pair_description(self):
        """Update port_pair: myid --description mydesc."""
        resource = 'port_pair'
        desc1 = "My_New_Port_Pair"
        cmd = pp.PortPairUpdate(test_cli20.MyApp(sys.stdout), None)
        self._test_update_resource(resource, cmd, 'myid',
                                   ['myid', '--description', desc1],
                                   {'description': desc1})

    def test_update_port_pair_name(self):
        """Update port_pair: myid --name myname."""
        resource = 'port_pair'
        my_name = "My_New_Port_Pair"
        cmd = pp.PortPairUpdate(test_cli20.MyApp(sys.stdout), None)
        self._test_update_resource(resource, cmd, 'myid',
                                   ['myid', '--name', my_name],
                                   {'name': my_name})

    def test_delete_port_pair(self):
        """Delete port-pair: myid."""
        resource = 'port_pair'
        cmd = pp.PortPairDelete(test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        args = [myid]
        self._test_delete_resource(resource, cmd, myid, args)

    def test_list_port_pair(self):
        """List port_pairs."""
        resources = 'port_pairs'
        cmd = pp.PortPairList(test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd, True)

    def test_list_port_pair_limit(self):
        """size (1000) limited list: port-pair -P."""
        resources = "port_pairs"
        cmd = pp.PortPairList(test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd, page_size=1000)

    def test_list_port_pairs_sort(self):
        """List port_pairs: --sort-key name --sort-key id --sort-key asc

        --sort-key desc
        """
        resources = "port_pairs"
        cmd = pp.PortPairList(test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd,
                                  sort_key=["name", "id"],
                                  sort_dir=["asc", "desc"])

    def test_show_port_pair(self):
        """Show port-pairs: --fields id --fields name myid."""
        resource = 'port_pair'
        cmd = pp.PortPairShow(test_cli20.MyApp(sys.stdout), None)
        args = ['--fields', 'id', '--fields', 'name', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id,
                                 args, ['id', 'name'])



# Copyright 2017 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from neutron.api.v2 import resource as api_res_log
from neutron import manager
from neutron.notifiers import nova as nova_log
from neutron.tests.unit.api.v2 import test_base as test_api_v2
from neutron.tests.unit.extensions import base as test_api_v2_extension
from oslo_config import cfg
from oslo_utils import uuidutils
from webob import exc
import webtest

from networking_sfc.extensions import sfc as sfc_ext

_uuid = uuidutils.generate_uuid
_get_path = test_api_v2._get_path

PORT_CHAIN_PATH = (sfc_ext.SFC_PREFIX[1:] + '/port_chains')
PORT_PAIR_PATH = (sfc_ext.SFC_PREFIX[1:] + '/port_pairs')
PORT_PAIR_GROUP_PATH = (sfc_ext.SFC_PREFIX[1:] + '/port_pair_groups')


class SfcExtensionTestCase(test_api_v2_extension.ExtensionTestCase):
    fmt = 'json'

    def setUp(self):
        self._mock_unnecessary_logging()
        super(SfcExtensionTestCase, self).setUp()
        self._setUpExtension(
            'networking_sfc.extensions.sfc.SfcPluginBase',
            sfc_ext.SFC_EXT,
            sfc_ext.RESOURCE_ATTRIBUTE_MAP,
            sfc_ext.Sfc,
            sfc_ext.SFC_PREFIX[1:],
            plural_mappings={}
        )

    def _mock_unnecessary_logging(self):
        mock_log_cfg_p = mock.patch.object(cfg, 'LOG')
        self.mock_log_cfg = mock_log_cfg_p.start()

        mock_log_manager_p = mock.patch.object(manager, 'LOG')
        self.mock_log_manager = mock_log_manager_p.start()

        mock_log_nova_p = mock.patch.object(nova_log, 'LOG')
        self.mock_log_nova = mock_log_nova_p.start()

        mock_log_api_res_log_p = mock.patch.object(api_res_log, 'LOG')
        self.mock_log_api_res_log = mock_log_api_res_log_p.start()

    @staticmethod
    def _get_expected_port_chain(data):
        port_chain = data['port_chain']
        chain_params = port_chain.get('chain_parameters') or dict()
        chain_params.setdefault('correlation', 'mpls')
        chain_params.setdefault('symmetric', False)
        ret = {'port_chain': {
            'description': port_chain.get('description') or '',
            'name': port_chain.get('name') or '',
            'port_pair_groups': port_chain['port_pair_groups'],
            'chain_parameters': chain_params,
            'flow_classifiers': port_chain.get(
                'flow_classifiers') or [],
            'tenant_id': port_chain['tenant_id'],
            'project_id': port_chain['project_id'],
            'chain_id': port_chain.get('chain_id') or 0
        }}
        return ret

    def _test_create_port_chain(self, **kwargs):
        tenant_id = _uuid()
        port_chain_data = {
            'port_pair_groups': [_uuid()],
            'tenant_id': tenant_id,
            'project_id': tenant_id
        }
        port_chain_data.update(kwargs)
        data = {'port_chain': port_chain_data}
        expected_data = self._get_expected_port_chain(data)
        return_value = copy.copy(expected_data['port_chain'])
        return_value.update({'id': _uuid()})
        instance = self.plugin.return_value
        instance.create_port_chain.return_value = return_value
        res = self.api.post(_get_path(PORT_CHAIN_PATH, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        instance.create_port_chain.assert_called_with(
            mock.ANY,
            port_chain=expected_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_chain', res)
        self.assertEqual(return_value, res['port_chain'])

    def test_create_port_chain(self):
        self._test_create_port_chain()

    def test_create_port_chain_all_fields(self):
        self._test_create_port_chain(description='desc',
                                     name='test1',
                                     chain_parameters={'symmetric': False,
                                                       'correlation': 'mpls'},
                                     flow_classifiers=[])

    def test_create_port_chain_all_fields_with_symmetric(self):
        self._test_create_port_chain(description='desc',
                                     name='test1',
                                     chain_parameters={'symmetric': True,
                                                       'correlation': 'mpls'},
                                     flow_classifiers=[])

    def test_create_port_chain_none_chain_parameters(self):
        self._test_create_port_chain(chain_parameters=None)

    def test_create_port_chain_empty_chain_parameters(self):
        self._test_create_port_chain(chain_parameters={})

    def test_create_port_chain_multiple_chain_parameters(self):
        self._test_create_port_chain(chain_parameters={
            'correlation': 'mpls',
            'symmetric': True
        })

    def test_create_port_chain_empty_port_pair_groups(self):
        tenant_id = _uuid()
        data = {'port_chain': {
            'port_pair_groups': [],
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_CHAIN_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_chain_nonuuid_port_pair_groups(self):
        tenant_id = _uuid()
        data = {'port_chain': {
            'port_pair_groups': ['nouuid'],
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_CHAIN_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_chain_nonuuid_flow_classifiers(self):
        tenant_id = _uuid()
        data = {'port_chain': {
            'port_pair_groups': [_uuid()],
            'flow_classifiers': ['nouuid'],
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_CHAIN_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_chain_invalid_chain_parameters(self):
        tenant_id = _uuid()
        data = {'port_chain': {
            'port_pair_groups': [_uuid()],
            'chain_parameters': {'abc': 'def'},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_CHAIN_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_chain_invalid_chain_parameters_correlation(self):
        tenant_id = _uuid()
        data = {'port_chain': {
            'port_pair_groups': [_uuid()],
            'chain_parameters': {'symmetric': False, 'correlation': 'def'},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_CHAIN_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_chain_invalid_chain_parameters_symmetric(self):
        tenant_id = _uuid()
        data = {'port_chain': {
            'port_pair_groups': [_uuid()],
            'chain_parameters': {'symmetric': 'abc', 'correlation': 'mpls'},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_CHAIN_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_port_chain_list(self):
        portchain_id = _uuid()
        tenant_id = _uuid()
        return_value = [{
            'tenant_id': tenant_id, 'project_id': tenant_id,
            'id': portchain_id
        }]
        instance = self.plugin.return_value
        instance.get_port_chains.return_value = return_value

        res = self.api.get(_get_path(PORT_CHAIN_PATH, fmt=self.fmt))

        instance.get_port_chains.assert_called_with(
            mock.ANY,
            fields=mock.ANY,
            filters=mock.ANY
        )
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_chains', res)
        self.assertEqual(return_value, res['port_chains'])

    def test_port_chain_get(self):
        portchain_id = _uuid()
        tenant_id = _uuid()
        return_value = {
            'tenant_id': tenant_id, 'project_id': tenant_id,
            'id': portchain_id
        }

        instance = self.plugin.return_value
        instance.get_port_chain.return_value = return_value

        res = self.api.get(_get_path(PORT_CHAIN_PATH,
                                     id=portchain_id, fmt=self.fmt))

        instance.get_port_chain.assert_called_with(
            mock.ANY,
            portchain_id,
            fields=mock.ANY
        )
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_chain', res)
        self.assertEqual(return_value, res['port_chain'])

    def test_port_chain_update(self):
        portchain_id = _uuid()
        tenant_id = _uuid()
        update_data = {'port_chain': {
            'name': 'new_name',
            'description': 'new_desc',
            'flow_classifiers': [_uuid()],
            'port_pair_groups': [_uuid()]
        }}
        return_value = {
            'tenant_id': tenant_id, 'project_id': tenant_id,
            'id': portchain_id
        }

        instance = self.plugin.return_value
        instance.update_port_chain.return_value = return_value

        res = self.api.put(_get_path(PORT_CHAIN_PATH, id=portchain_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        instance.update_port_chain.assert_called_with(
            mock.ANY, portchain_id,
            port_chain=update_data)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_chain', res)
        self.assertEqual(return_value, res['port_chain'])

    def test_port_chain_update_nonuuid_flow_classifiers(self):
        portchain_id = _uuid()
        data = {'port_chain': {
            'flow_classifiers': ['nouuid'],
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.put,
            _get_path(PORT_CHAIN_PATH, id=portchain_id, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_port_chain_update_nonuuid_port_pair_groups(self):
        portchain_id = _uuid()
        update_data = {'port_chain': {
            'port_pair_groups': ['nouuid']
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.put,
            _get_path(PORT_CHAIN_PATH, id=portchain_id,
                      fmt=self.fmt),
            self.serialize(update_data),
            content_type='application/%s' % self.fmt
        )

    def test_port_chain_update_chain_parameters(self):
        portchain_id = _uuid()
        update_data = {'port_chain': {
            'chain_parameters': {}
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.put,
            _get_path(PORT_CHAIN_PATH, id=portchain_id,
                      fmt=self.fmt),
            self.serialize(update_data)
        )

    def test_port_chain_delete(self):
        self._test_entity_delete('port_chain')

    def _get_expected_port_pair_group(self, data):
        port_pair_group = data['port_pair_group']
        ret = {'port_pair_group': {
            'description': port_pair_group.get('description') or '',
            'name': port_pair_group.get('name') or '',
            'port_pairs': port_pair_group.get('port_pairs') or [],
            'tenant_id': port_pair_group['tenant_id'],
            'project_id': port_pair_group['project_id'],
            'port_pair_group_parameters': port_pair_group.get(
                'port_pair_group_parameters'
            ) or {'lb_fields': [],
                  'ppg_n_tuple_mapping': {'ingress_n_tuple': {},
                                          'egress_n_tuple': {}}
                  }
        }}
        if port_pair_group.get('group_id'):
            ret['port_pair_group']['group_id'] = port_pair_group['group_id']
        return ret

    def test_create_port_pair_group(self):
        portpairgroup_id = _uuid()
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        expected_data = self._get_expected_port_pair_group(data)
        return_value = copy.copy(expected_data['port_pair_group'])
        return_value.update({'id': portpairgroup_id})
        instance = self.plugin.return_value
        instance.create_port_pair_group.return_value = return_value
        res = self.api.post(
            _get_path(PORT_PAIR_GROUP_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)
        instance.create_port_pair_group.assert_called_with(
            mock.ANY,
            port_pair_group=expected_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair_group', res)
        self.assertEqual(return_value, res['port_pair_group'])

    def test_create_port_pair_group_all_fields(self):
        portpairgroup_id = _uuid()
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'description': 'desc',
            'name': 'test1',
            'port_pairs': [],
            'port_pair_group_parameters': {},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        expected_data = self._get_expected_port_pair_group(data)
        return_value = copy.copy(expected_data['port_pair_group'])
        return_value.update({'id': portpairgroup_id})
        instance = self.plugin.return_value
        instance.create_port_pair_group.return_value = return_value
        res = self.api.post(
            _get_path(PORT_PAIR_GROUP_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)
        instance.create_port_pair_group.assert_called_with(
            mock.ANY,
            port_pair_group=expected_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair_group', res)
        self.assertEqual(return_value, res['port_pair_group'])

    def test_create_port_pair_group_none_parameters(self):
        portpairgroup_id = _uuid()
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'port_pairs': [_uuid()],
            'port_pair_group_parameters': None,
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        expected_data = self._get_expected_port_pair_group(data)
        return_value = copy.copy(expected_data['port_pair_group'])
        return_value.update({'id': portpairgroup_id})
        instance = self.plugin.return_value
        instance.create_port_pair_group.return_value = return_value
        res = self.api.post(_get_path(PORT_PAIR_GROUP_PATH, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        instance.create_port_pair_group.assert_called_with(
            mock.ANY,
            port_pair_group=expected_data)
        self.assertEqual(res.status_int, exc.HTTPCreated.code)
        res = self.deserialize(res)
        self.assertIn('port_pair_group', res)
        self.assertEqual(return_value, res['port_pair_group'])

    def test_create_port_pair_group_empty_parameters(self):
        portpairgroup_id = _uuid()
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'port_pairs': [_uuid()],
            'port_pair_group_parameters': {},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        expected_data = self._get_expected_port_pair_group(data)
        return_value = copy.copy(expected_data['port_pair_group'])
        return_value.update({'id': portpairgroup_id})
        instance = self.plugin.return_value
        instance.create_port_pair_group.return_value = return_value
        res = self.api.post(_get_path(PORT_PAIR_GROUP_PATH, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        instance.create_port_pair_group.assert_called_with(
            mock.ANY,
            port_pair_group=expected_data)
        self.assertEqual(res.status_int, exc.HTTPCreated.code)
        res = self.deserialize(res)
        self.assertIn('port_pair_group', res)
        self.assertEqual(return_value, res['port_pair_group'])

    def test_create_port_pair_group_invalid_parameters(self):
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'port_pairs': [_uuid()],
            'port_pair_group_parameters': {'abc': 'def'},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_PAIR_GROUP_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_pair_group_invalid_lb_fields_type(self):
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'port_pairs': [_uuid()],
            'port_pair_group_parameters': {'lb_fields': 'ip_src'},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_PAIR_GROUP_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_pair_group_invalid_lb_fields(self):
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'port_pairs': [_uuid()],
            'port_pair_group_parameters': {'lb_fields': ['def']},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_PAIR_GROUP_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_pair_group_invalid_ppg_n_tuple_mapping_key(self):
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'port_pairs': [_uuid()],
            'port_pair_group_parameters': {
                'ppg_n_tuple_mapping': {
                    'ingress_n_tuple': {'sssource_ip_prefix': None},
                    'egress_n_tuple': {'protool': None}
                }
            },
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_PAIR_GROUP_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_pair_group_nonuuid_port_pairs(self):
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'port_pairs': ['nouuid'],
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_PAIR_GROUP_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_port_pair_group_list(self):
        portpairgroup_id = _uuid()
        tenant_id = _uuid()
        return_value = [{
            'tenant_id': tenant_id, 'project_id': tenant_id,
            'id': portpairgroup_id
        }]
        instance = self.plugin.return_value
        instance.get_port_pair_groups.return_value = return_value

        res = self.api.get(
            _get_path(PORT_PAIR_GROUP_PATH, fmt=self.fmt))

        instance.get_port_pair_groups.assert_called_with(
            mock.ANY,
            fields=mock.ANY,
            filters=mock.ANY
        )
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair_groups', res)
        self.assertEqual(return_value, res['port_pair_groups'])

    def test_port_pair_group_get(self):
        portpairgroup_id = _uuid()
        tenant_id = _uuid()
        return_value = {
            'tenant_id': tenant_id, 'project_id': tenant_id,
            'id': portpairgroup_id
        }

        instance = self.plugin.return_value
        instance.get_port_pair_group.return_value = return_value

        res = self.api.get(_get_path(PORT_PAIR_GROUP_PATH,
                                     id=portpairgroup_id, fmt=self.fmt))

        instance.get_port_pair_group.assert_called_with(
            mock.ANY,
            portpairgroup_id,
            fields=mock.ANY
        )
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair_group', res)
        self.assertEqual(return_value, res['port_pair_group'])

    def test_port_pair_group_update(self):
        portpairgroup_id = _uuid()
        tenant_id = _uuid()
        update_data = {'port_pair_group': {
            'name': 'new_name',
            'description': 'new_desc',
            'port_pairs': [_uuid()]
        }}
        return_value = {
            'tenant_id': tenant_id, 'project_id': tenant_id,
            'id': portpairgroup_id
        }

        instance = self.plugin.return_value
        instance.update_port_pair_group.return_value = return_value

        res = self.api.put(
            _get_path(
                PORT_PAIR_GROUP_PATH, id=portpairgroup_id,
                fmt=self.fmt),
            self.serialize(update_data))

        instance.update_port_pair_group.assert_called_with(
            mock.ANY, portpairgroup_id,
            port_pair_group=update_data)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair_group', res)
        self.assertEqual(return_value, res['port_pair_group'])

    def test_port_pair_group_update_nonuuid_port_pairs(self):
        portpairgroup_id = _uuid()
        data = {'port_pair_group': {
            'port_pairs': ['nouuid']
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.put,
            _get_path(PORT_PAIR_GROUP_PATH, id=portpairgroup_id,
                      fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_port_pair_group_delete(self):
        self._test_entity_delete('port_pair_group')

    def _get_expected_port_pair(self, data):
        return {'port_pair': {
            'name': data['port_pair'].get('name') or '',
            'description': data['port_pair'].get('description') or '',
            'ingress': data['port_pair']['ingress'],
            'egress': data['port_pair']['egress'],
            'service_function_parameters': data['port_pair'].get(
                'service_function_parameters') or {
                'correlation': None, 'weight': 1},
            'tenant_id': data['port_pair']['tenant_id'],
            'project_id': data['port_pair']['project_id']
        }}

    def test_create_port_pair(self):
        portpair_id = _uuid()
        tenant_id = _uuid()
        data = {'port_pair': {
            'ingress': _uuid(),
            'egress': _uuid(),
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        expected_data = self._get_expected_port_pair(data)
        return_value = copy.copy(expected_data['port_pair'])
        return_value.update({'id': portpair_id})
        instance = self.plugin.return_value
        instance.create_port_pair.return_value = return_value
        res = self.api.post(_get_path(PORT_PAIR_PATH, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        instance.create_port_pair.assert_called_with(
            mock.ANY,
            port_pair=expected_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair', res)
        self.assertEqual(return_value, res['port_pair'])

    def test_create_port_pair_all_fields(self):
        portpair_id = _uuid()
        tenant_id = _uuid()
        data = {'port_pair': {
            'description': 'desc',
            'name': 'test1',
            'ingress': _uuid(),
            'egress': _uuid(),
            'service_function_parameters': {
                'correlation': None, 'weight': 2},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        expected_data = self._get_expected_port_pair(data)
        return_value = copy.copy(expected_data['port_pair'])
        return_value.update({'id': portpair_id})
        instance = self.plugin.return_value
        instance.create_port_pair.return_value = return_value
        res = self.api.post(_get_path(PORT_PAIR_PATH, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        instance.create_port_pair.assert_called_with(
            mock.ANY,
            port_pair=expected_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair', res)
        self.assertEqual(return_value, res['port_pair'])

    def test_create_port_pair_non_service_function_parameters(self):
        portpair_id = _uuid()
        tenant_id = _uuid()
        data = {'port_pair': {
            'ingress': _uuid(),
            'egress': _uuid(),
            'service_function_parameters': None,
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        expected_data = self._get_expected_port_pair(data)
        return_value = copy.copy(expected_data['port_pair'])
        return_value.update({'id': portpair_id})
        instance = self.plugin.return_value
        instance.create_port_pair.return_value = return_value
        res = self.api.post(_get_path(PORT_PAIR_PATH, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        instance.create_port_pair.assert_called_with(
            mock.ANY,
            port_pair=expected_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair', res)
        self.assertEqual(return_value, res['port_pair'])

    def test_create_port_pair_empty_service_function_parameters(self):
        portpair_id = _uuid()
        tenant_id = _uuid()
        data = {'port_pair': {
            'ingress': _uuid(),
            'egress': _uuid(),
            'service_function_parameters': {},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        expected_data = self._get_expected_port_pair(data)
        return_value = copy.copy(expected_data['port_pair'])
        return_value.update({'id': portpair_id})
        instance = self.plugin.return_value
        instance.create_port_pair.return_value = return_value
        res = self.api.post(_get_path(PORT_PAIR_PATH, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        instance.create_port_pair.assert_called_with(
            mock.ANY,
            port_pair=expected_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair', res)
        self.assertEqual(return_value, res['port_pair'])

    def test_create_port_pair_invalid_service_function_parameters(self):
        tenant_id = _uuid()
        data = {'port_pair': {
            'ingress': _uuid(),
            'egress': _uuid(),
            'service_function_parameters': {'abc': 'def'},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_PAIR_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_pair_invalid_correlation(self):
        tenant_id = _uuid()
        data = {'port_pair': {
            'ingress': _uuid(),
            'egress': _uuid(),
            'service_function_parameters': {'correlation': 'def'},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_PAIR_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_pair_invalid_weight_type(self):
        tenant_id = _uuid()
        data = {'port_pair': {
            'ingress': _uuid(),
            'egress': _uuid(),
            'service_function_parameters': {'weight': 'abc'},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_PAIR_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_pair_invalid_weight(self):
        tenant_id = _uuid()
        data = {'port_pair': {
            'ingress': _uuid(),
            'egress': _uuid(),
            'service_function_parameters': {'weight': -1},
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_PAIR_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_pair_nouuid_ingress(self):
        tenant_id = _uuid()
        data = {'port_pair': {
            'ingress': 'abc',
            'egress': _uuid(),
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_PAIR_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_create_port_pair_nouuid_egress(self):
        tenant_id = _uuid()
        data = {'port_pair': {
            'egress': 'abc',
            'ingress': _uuid(),
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(PORT_PAIR_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_port_pair_list(self):
        portpair_id = _uuid()
        tenant_id = _uuid()
        return_value = [{
            'tenant_id': tenant_id, 'project_id': tenant_id,
            'id': portpair_id
        }]
        instance = self.plugin.return_value
        instance.get_port_pairs.return_value = return_value

        res = self.api.get(_get_path(PORT_PAIR_PATH, fmt=self.fmt))

        instance.get_port_pairs.assert_called_with(
            mock.ANY,
            fields=mock.ANY,
            filters=mock.ANY
        )
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pairs', res)
        self.assertEqual(return_value, res['port_pairs'])

    def test_port_pair_get(self):
        portpair_id = _uuid()
        tenant_id = _uuid()
        return_value = {
            'tenant_id': tenant_id, 'project_id': tenant_id,
            'id': portpair_id
        }

        instance = self.plugin.return_value
        instance.get_port_pair.return_value = return_value

        res = self.api.get(_get_path(PORT_PAIR_PATH,
                                     id=portpair_id, fmt=self.fmt))

        instance.get_port_pair.assert_called_with(
            mock.ANY,
            portpair_id,
            fields=mock.ANY
        )
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair', res)
        self.assertEqual(return_value, res['port_pair'])

    def test_port_pair_update(self):
        portpair_id = _uuid()
        tenant_id = _uuid()
        update_data = {'port_pair': {
            'name': 'new_name',
            'description': 'new_desc'
        }}
        return_value = {
            'tenant_id': tenant_id, 'project_id': tenant_id,
            'id': portpair_id
        }

        instance = self.plugin.return_value
        instance.update_port_pair.return_value = return_value

        res = self.api.put(_get_path(PORT_PAIR_PATH, id=portpair_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        instance.update_port_pair.assert_called_with(
            mock.ANY, portpair_id,
            port_pair=update_data)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair', res)
        self.assertEqual(return_value, res['port_pair'])

    def test_port_pair_update_service_function_parameters(self):
        portpair_id = _uuid()
        data = {'port_pair': {
            'service_function_parameters': None
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.put,
            _get_path(PORT_PAIR_PATH, id=portpair_id,
                      fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_port_pair_update_ingress(self):
        portpair_id = _uuid()
        data = {'port_pair': {
            'ingress': _uuid()
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.put,
            _get_path(PORT_PAIR_PATH, id=portpair_id,
                      fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_port_pair_update_egress(self):
        portpair_id = _uuid()
        data = {'port_pair': {
            'egress': _uuid()
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.put,
            _get_path(PORT_PAIR_PATH, id=portpair_id,
                      fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_port_pair_delete(self):
        self._test_entity_delete('port_pair')

    # NOTE(scsnow): move to neutron-lib
    def test_validate_list_of_allowed_values(self):
        data = ['eth_src', 'eth_src', 'illegal']
        allowed_values = ['eth_src', 'eth_src']
        msg = sfc_ext.validate_list_of_allowed_values(data, allowed_values)
        self.assertIn("Illegal values in a list:", msg)



# Copyright 2017 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutron.db import api as db_api
from neutron.db import common_db_mixin
from neutron_lib import context as n_context
from neutron_lib.db import model_base
from neutron_lib import exceptions as n_exc

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import exc
from sqlalchemy import sql

import six

from oslo_log import helpers as log_helpers
from oslo_utils import uuidutils

from networking_sfc._i18n import _


class PortPairDetailNotFound(n_exc.NotFound):
    message = _("Portchain port brief %(port_id)s could not be found")


class NodeNotFound(n_exc.NotFound):
    message = _("Portchain node %(node_id)s could not be found")


# name changed to ChainPathId
class UuidIntidAssoc(model_base.BASEV2, model_base.HasId):
    __tablename__ = 'sfc_uuid_intid_associations'
    uuid = sa.Column(sa.String(36), primary_key=True)
    intid = sa.Column(sa.Integer, unique=True, nullable=False)
    type_ = sa.Column(sa.String(32), nullable=False)

    def __init__(self, uuid, intid, type_):
        self.uuid = uuid
        self.intid = intid
        self.type_ = type_


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return getinstance


@singleton
class IDAllocation(object):
    def __init__(self, context):
        # Get the initial range from conf file.
        conf_obj = {'group': [1, 255], 'portchain': [256, 65536]}
        self.conf_obj = conf_obj
        self.context = context

    @log_helpers.log_method_call
    def assign_intid(self, type_, uuid):
        query = self.context.session.query(UuidIntidAssoc).filter_by(
            type_=type_).order_by(UuidIntidAssoc.intid)

        allocated_int_ids = {obj.intid for obj in query.all()}

        # Find the first one from the available range that
        # is not in allocated_int_ids
        start, end = self.conf_obj[type_][0], self.conf_obj[type_][1] + 1
        for init_id in six.moves.range(start, end):
            if init_id not in allocated_int_ids:
                with db_api.context_manager.writer.using(self.context):
                    uuid_intid = UuidIntidAssoc(
                        uuid, init_id, type_)
                    self.context.session.add(uuid_intid)
                return init_id
        return None

    @log_helpers.log_method_call
    def get_intid_by_uuid(self, type_, uuid):

        query_obj = self.context.session.query(UuidIntidAssoc).filter_by(
            type_=type_, uuid=uuid).first()
        if query_obj:
            return query_obj.intid
        return None

    @log_helpers.log_method_call
    def release_intid(self, type_, intid):
        """Release int id.

        @param: type_: str
        @param: intid: int
        """
        with db_api.context_manager.writer.using(self.context):
            query_obj = self.context.session.query(UuidIntidAssoc).filter_by(
                intid=intid, type_=type_).first()

            if query_obj:
                self.session.delete(query_obj)


class PathPortAssoc(model_base.BASEV2):
    """path port association table.

    It represents the association table which associate path_nodes with
    portpair_details.
    """
    __tablename__ = 'sfc_path_port_associations'
    pathnode_id = sa.Column(sa.String(36),
                            sa.ForeignKey(
                                'sfc_path_nodes.id', ondelete='CASCADE'),
                            primary_key=True)
    portpair_id = sa.Column(sa.String(36),
                            sa.ForeignKey('sfc_portpair_details.id',
                                          ondelete='CASCADE'),
                            primary_key=True)
    weight = sa.Column(sa.Integer, nullable=False, default=1)


class PortPairDetail(model_base.BASEV2, model_base.HasId,
                     model_base.HasProject):
    __tablename__ = 'sfc_portpair_details'
    ingress = sa.Column(sa.String(36), nullable=True)
    egress = sa.Column(sa.String(36), nullable=True)
    host_id = sa.Column(sa.String(255), nullable=False)
    in_mac_address = sa.Column(sa.String(32))
    mac_address = sa.Column(sa.String(32), nullable=False)
    network_type = sa.Column(sa.String(8))
    segment_id = sa.Column(sa.Integer)
    local_endpoint = sa.Column(sa.String(64), nullable=False)
    path_nodes = orm.relationship(PathPortAssoc,
                                  backref='port_pair_detail',
                                  lazy="joined",
                                  cascade='all,delete')
    correlation = sa.Column(sa.String(255), nullable=True)


class PathNode(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    __tablename__ = 'sfc_path_nodes'
    nsp = sa.Column(sa.Integer, nullable=False)
    nsi = sa.Column(sa.Integer, nullable=False)
    node_type = sa.Column(sa.String(32))
    portchain_id = sa.Column(
        sa.String(255),
        sa.ForeignKey('sfc_port_chains.id', ondelete='CASCADE'))
    status = sa.Column(sa.String(32))
    portpair_details = orm.relationship(PathPortAssoc,
                                        backref='path_nodes',
                                        lazy="joined",
                                        cascade='all,delete')
    next_group_id = sa.Column(sa.Integer)
    next_hop = sa.Column(sa.String(512))
    fwd_path = sa.Column(sa.Boolean(),
                         nullable=False)
    ppg_n_tuple_mapping = sa.Column(sa.String(1024), nullable=True)
    tap_enabled = sa.Column(sa.Boolean(),
                            nullable=False,
                            server_default=sa.sql.false())
    previous_node_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('sfc_path_nodes.id', ondelete='SET NULL'))


class OVSSfcDriverDB(common_db_mixin.CommonDbMixin):
    def initialize(self):
        self.admin_context = n_context.get_admin_context()

    def _make_pathnode_dict(self, node, fields=None):
        res = {'id': node['id'],
               'project_id': node['project_id'],
               'node_type': node['node_type'],
               'nsp': node['nsp'],
               'nsi': node['nsi'],
               'next_group_id': node['next_group_id'],
               'next_hop': node['next_hop'],
               'portchain_id': node['portchain_id'],
               'status': node['status'],
               'portpair_details': [pair_detail['portpair_id']
                                    for pair_detail in node['portpair_details']
                                    ],
               'fwd_path': node['fwd_path'],
               'ppg_n_tuple_mapping': node['ppg_n_tuple_mapping'],
               'tap_enabled': node['tap_enabled'],
               'previous_node_id': node['previous_node_id']
               }

        return self._fields(res, fields)

    def _make_port_detail_dict(self, port, fields=None):
        res = {'id': port['id'],
               'project_id': port['project_id'],
               'host_id': port['host_id'],
               'ingress': port.get('ingress', None),
               'egress': port.get('egress', None),
               'segment_id': port['segment_id'],
               'local_endpoint': port['local_endpoint'],
               'mac_address': port['mac_address'],
               'in_mac_address': port['in_mac_address'],
               'network_type': port['network_type'],
               'path_nodes': [{'pathnode_id': node['pathnode_id'],
                               'weight': node['weight']}
                              for node in port['path_nodes']],
               'correlation': port['correlation']
               }

        return self._fields(res, fields)

    def _make_pathport_assoc_dict(self, assoc, fields=None):
        res = {'pathnode_id': assoc['pathnode_id'],
               'portpair_id': assoc['portpair_id'],
               'weight': assoc['weight'],
               }

        return self._fields(res, fields)

    def _get_path_node(self, id):
        try:
            node = self._get_by_id(self.admin_context, PathNode, id)
        except exc.NoResultFound:
            raise NodeNotFound(node_id=id)
        return node

    def _get_port_pair_detail(self, id):
        try:
            port = self._get_by_id(self.admin_context, PortPairDetail, id)
        except exc.NoResultFound:
            raise PortPairDetailNotFound(port_id=id)
        return port

    def create_port_pair_detail(self, port):
        with db_api.context_manager.writer.using(self.admin_context):
            args = self._filter_non_model_columns(port, PortPairDetail)
            args['id'] = uuidutils.generate_uuid()
            port_obj = PortPairDetail(**args)
            self.admin_context.session.add(port_obj)
            return self._make_port_detail_dict(port_obj)

    def create_path_node(self, node):
        with db_api.context_manager.writer.using(self.admin_context):
            args = self._filter_non_model_columns(node, PathNode)
            args['id'] = uuidutils.generate_uuid()
            node_obj = PathNode(**args)
            self.admin_context.session.add(node_obj)
            return self._make_pathnode_dict(node_obj)

    def create_pathport_assoc(self, assoc):
        with db_api.context_manager.writer.using(self.admin_context):
            args = self._filter_non_model_columns(assoc, PathPortAssoc)
            assoc_obj = PathPortAssoc(**args)
            self.admin_context.session.add(assoc_obj)
            return self._make_pathport_assoc_dict(assoc_obj)

    def delete_pathport_assoc(self, pathnode_id, portdetail_id):
        with db_api.context_manager.writer.using(self.admin_context):
            self.admin_context.session.query(PathPortAssoc).filter_by(
                pathnode_id=pathnode_id,
                portpair_id=portdetail_id).delete()

    def update_port_detail(self, id, port):
        with db_api.context_manager.writer.using(self.admin_context):
            port_obj = self._get_port_detail(id)
            for key, value in port.items():
                if key == 'path_nodes':
                    pns = []
                    for pn in value:
                        pn_id = pn['pathnode_id']
                        self._get_path_node(pn_id)
                        query = self._model_query(
                            self.admin_context, PathPortAssoc)
                        pn_association = query.filter_by(
                            pathnode_id=pn_id,
                            portpair_id=id
                        ).first()
                        if not pn_association:
                            pn_association = PathPortAssoc(
                                pathnode_id=pn_id,
                                portpair_id=id,
                                weight=pn.get('weight', 1)
                            )
                        pns.append(pn_association)
                    port_obj[key] = pns
                else:
                    port_obj[key] = value
            port_obj.update(port)
            return self._make_port_detail_dict(port_obj)

    def update_path_node(self, id, node):
        with db_api.context_manager.writer.using(self.admin_context):
            node_obj = self._get_path_node(id)
            for key, value in node.items():
                if key == 'portpair_details':
                    pds = []
                    for pd_id in value:
                        query = self._model_query(
                            self.admin_context, PathPortAssoc)
                        pd_association = query.filter_by(
                            pathnode_id=id,
                            portpair_id=pd_id
                        ).first()
                        if not pd_association:
                            pd_association = PathPortAssoc(
                                pathnode_id=id,
                                portpair_id=pd_id
                            )
                        pds.append(pd_association)
                    node_obj[key] = pds
                else:
                    node_obj[key] = value

            return self._make_pathnode_dict(node_obj)

    def delete_port_pair_detail(self, id):
        with db_api.context_manager.writer.using(self.admin_context):
            port_obj = self._get_port_pair_detail(id)
            self.admin_context.session.delete(port_obj)

    def delete_path_node(self, id):
        with db_api.context_manager.writer.using(self.admin_context):
            node_obj = self._get_path_node(id)
            self.admin_context.session.delete(node_obj)

    def get_port_detail(self, id):
        with db_api.context_manager.reader.using(self.admin_context):
            port_obj = self._get_port_pair_detail(id)
            return self._make_port_detail_dict(port_obj)

    def get_port_detail_without_exception(self, id):
        with db_api.context_manager.reader.using(self.admin_context):
            try:
                port = self._get_by_id(
                    self.admin_context, PortPairDetail, id)
            except exc.NoResultFound:
                return None
            return self._make_port_detail_dict(port)

    def get_path_node(self, id):
        with db_api.context_manager.reader.using(self.admin_context):
            node_obj = self._get_path_node(id)
        return self._make_pathnode_dict(node_obj)

    def get_path_nodes_by_filter(self, filters=None):
        with db_api.context_manager.reader.using(self.admin_context):
            qry = self._get_path_nodes_by_filter(filters)
            all_items = qry.all()
            if all_items:
                return [self._make_pathnode_dict(item) for item in all_items]
        return None

    def get_path_node_by_filter(self, filters=None):
        with db_api.context_manager.reader.using(self.admin_context):
            qry = self._get_path_nodes_by_filter(filters)
            first = qry.first()
            if first:
                return self._make_pathnode_dict(first)
        return None

    def _get_path_nodes_by_filter(self, filters=None):
        qry = self.admin_context.session.query(PathNode)
        if filters:
            for key, value in filters.items():
                column = getattr(PathNode, key, None)
                if column:
                    if not value:
                        qry = qry.filter(sql.false())
                    else:
                        qry = qry.filter(column == value)
        return qry

    def get_port_details_by_filter(self, filters=None):
        with db_api.context_manager.reader.using(self.admin_context):
            qry = self._get_port_details_by_filter(filters)
            all_items = qry.all()
            if all_items:
                return [self._make_port_detail_dict(item)
                        for item in all_items]
        return None

    def get_port_detail_by_filter(self, filters=None):
        with db_api.context_manager.reader.using(self.admin_context):
            qry = self._get_port_details_by_filter(filters)
            first = qry.first()
            if first:
                return self._make_port_detail_dict(first)
        return None

    def _get_port_details_by_filter(self, filters=None):
        qry = self.admin_context.session.query(PortPairDetail)
        if filters:
            for key, value in filters.items():
                column = getattr(PortPairDetail, key, None)
                if column:
                    if not value:
                        qry = qry.filter(sql.false())
                    else:
                        qry = qry.filter(column == value)
        return qry



# Copyright (c) 2017 One Convergence Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_sfc.extensions import sfc
from neutron_lib.api import converters as lib_converters
from neutron_lib.api import extensions
from neutron_lib import exceptions as neutron_exc


DEFAULT_TAP_ENABLED = False
TAP_EXT = "networking-sfc-tap"


class MultiplePortPairsInTapPPGNotSupported(neutron_exc.InvalidInput):
    """Multiple Port Pairs in Tap PPG not allowed"""
    message = _("Multiple port pairs in Tap port-pair-group is not allowed.")


class ConsecutiveTapPPGNotSupported(neutron_exc.InvalidInput):
    """Unsupported Tap deployment"""
    message = _("Consecutive Tap PPG in port-chain is not supported now.")


EXTENDED_ATTRIBUTES_2_0 = {
    'port_pair_groups': {
        'tap_enabled': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True, 'default': DEFAULT_TAP_ENABLED,
            'convert_to': lib_converters.convert_to_boolean
        }
    }
}


class Tap(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Networking SFC Passive Tap Service Function support"

    @classmethod
    def get_alias(cls):
        return TAP_EXT

    @classmethod
    def get_description(cls):
        return "Extension for Passive TAP Service Function support"

    @classmethod
    def get_updated(cls):
        return "2017-10-20T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}

    @classmethod
    def get_plugin_interface(cls):
        return sfc.SfcPluginBase



# Copyright 2015 Futurewei.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import uuidutils

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.orm import exc

from neutron_lib import constants as const
from neutron_lib.db import model_base

from neutron.db import api as db_api
from neutron.db import common_db_mixin
from neutron.db import models_v2

from networking_sfc.extensions import flowclassifier as fc_ext

LOG = logging.getLogger(__name__)
UUID_LEN = 36


class L7Parameter(model_base.BASEV2):
    """Represents a L7 parameter."""
    __tablename__ = 'sfc_flow_classifier_l7_parameters'
    keyword = sa.Column(sa.String(255), primary_key=True)
    value = sa.Column(sa.String(255))
    classifier_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('sfc_flow_classifiers.id', ondelete='CASCADE'),
        primary_key=True)


class FlowClassifier(model_base.BASEV2, model_base.HasId,
                     model_base.HasProject):
    """Represents a v2 neutron flow classifier."""
    __tablename__ = 'sfc_flow_classifiers'
    name = sa.Column(sa.String(255))
    ethertype = sa.Column(sa.String(40))
    protocol = sa.Column(sa.String(40))
    description = sa.Column(sa.String(255))
    source_port_range_min = sa.Column(sa.Integer)
    source_port_range_max = sa.Column(sa.Integer)
    destination_port_range_min = sa.Column(sa.Integer)
    destination_port_range_max = sa.Column(sa.Integer)
    source_ip_prefix = sa.Column(sa.String(255))
    destination_ip_prefix = sa.Column(sa.String(255))
    logical_source_port = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('ports.id', ondelete='RESTRICT'))
    logical_destination_port = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('ports.id', ondelete='RESTRICT'))
    l7_parameters = orm.relationship(
        L7Parameter,
        collection_class=attribute_mapped_collection('keyword'),
        cascade='all, delete-orphan')


class FlowClassifierDbPlugin(fc_ext.FlowClassifierPluginBase,
                             common_db_mixin.CommonDbMixin):

    @classmethod
    def _check_port_range_valid(cls, port_range_min,
                                port_range_max,
                                protocol):
        if (
            port_range_min is not None and
            port_range_max is not None and
            port_range_min > port_range_max
        ):
            raise fc_ext.FlowClassifierInvalidPortRange(
                port_range_min=port_range_min,
                port_range_max=port_range_max
            )
        if port_range_min is not None or port_range_max is not None:
            if protocol not in [const.PROTO_NAME_TCP, const.PROTO_NAME_UDP]:
                raise fc_ext.FlowClassifierProtocolRequiredWithPorts()

    @classmethod
    def _check_ip_prefix_valid(cls, ip_prefix, ethertype):
        if ip_prefix is not None:
            ip = netaddr.IPNetwork(ip_prefix)
            if ethertype == 'IPv4' and ip.version == 4:
                pass
            elif ethertype == 'IPv6' and ip.version == 6:
                pass
            else:
                raise (
                    fc_ext.FlowClassifierIpPrefixFormatConflictWithEthertype(
                        ip_prefix=ip_prefix, ethertype=ethertype
                    )
                )

    @classmethod
    def _logical_port_conflict(cls, first_logical_port, second_logical_port):
        if first_logical_port is None or second_logical_port is None:
            return True
        return first_logical_port == second_logical_port

    @classmethod
    def _ip_prefix_conflict(cls, first_ip_prefix, second_ip_prefix):
        if first_ip_prefix is None or second_ip_prefix is None:
            return True
        first_ipset = netaddr.IPSet([first_ip_prefix])
        second_ipset = netaddr.IPSet([second_ip_prefix])
        return bool(first_ipset & second_ipset)

    @classmethod
    def _port_range_conflict(
        cls, first_port_range_min, first_port_range_max,
        second_port_range_min, second_port_range_max
    ):
        first_conflict = True
        second_conflict = True
        if (
            first_port_range_min is not None and
            second_port_range_max is not None
        ):
            first_conflict = first_port_range_min <= second_port_range_max
        if (
            first_port_range_max is not None and
            second_port_range_min is not None
        ):
            second_conflict = second_port_range_min <= first_port_range_max
        return first_conflict & second_conflict

    @classmethod
    def _protocol_conflict(cls, first_protocol, second_protocol):
        if first_protocol is None or second_protocol is None:
            return True
        return first_protocol == second_protocol

    @classmethod
    def _ethertype_conflict(cls, first_ethertype, second_ethertype):
        return first_ethertype == second_ethertype

    @classmethod
    def flowclassifier_basic_conflict(
        cls, first_flowclassifier, second_flowclassifier
    ):
        return all([
            cls._ethertype_conflict(
                first_flowclassifier['ethertype'],
                second_flowclassifier['ethertype']
            ),
            cls._protocol_conflict(
                first_flowclassifier['protocol'],
                second_flowclassifier['protocol']
            ),
            cls._ip_prefix_conflict(
                first_flowclassifier['source_ip_prefix'],
                second_flowclassifier['source_ip_prefix']
            ),
            cls._ip_prefix_conflict(
                first_flowclassifier['destination_ip_prefix'],
                second_flowclassifier['destination_ip_prefix']
            ),
            cls._port_range_conflict(
                first_flowclassifier['source_port_range_min'],
                first_flowclassifier['source_port_range_max'],
                second_flowclassifier['source_port_range_min'],
                second_flowclassifier['source_port_range_max']
            ),
            cls._port_range_conflict(
                first_flowclassifier['destination_port_range_min'],
                first_flowclassifier['destination_port_range_max'],
                second_flowclassifier['destination_port_range_min'],
                second_flowclassifier['destination_port_range_max']
            )
        ])

    @classmethod
    def flowclassifier_conflict(
        cls, first_flowclassifier, second_flowclassifier
    ):
        return all([
            cls.flowclassifier_basic_conflict(
                first_flowclassifier,
                second_flowclassifier
            ),
            cls._logical_port_conflict(
                first_flowclassifier['logical_source_port'],
                second_flowclassifier['logical_source_port']
            ),
            cls._logical_port_conflict(
                first_flowclassifier['logical_destination_port'],
                second_flowclassifier['logical_destination_port']
            )
        ])

    @log_helpers.log_method_call
    def create_flow_classifier(self, context, flow_classifier):
        fc = flow_classifier['flow_classifier']
        project_id = fc['project_id']
        l7_parameters = {
            key: L7Parameter(key, val)
            for key, val in fc['l7_parameters'].items()}
        ethertype = fc['ethertype']
        protocol = fc['protocol']
        source_port_range_min = fc['source_port_range_min']
        source_port_range_max = fc['source_port_range_max']
        self._check_port_range_valid(source_port_range_min,
                                     source_port_range_max,
                                     protocol)
        destination_port_range_min = fc['destination_port_range_min']
        destination_port_range_max = fc['destination_port_range_max']
        self._check_port_range_valid(destination_port_range_min,
                                     destination_port_range_max,
                                     protocol)
        source_ip_prefix = fc['source_ip_prefix']
        self._check_ip_prefix_valid(source_ip_prefix, ethertype)
        destination_ip_prefix = fc['destination_ip_prefix']
        self._check_ip_prefix_valid(destination_ip_prefix, ethertype)
        logical_source_port = fc['logical_source_port']
        logical_destination_port = fc['logical_destination_port']
        with db_api.context_manager.writer.using(context):
            if logical_source_port is not None:
                self._get_port(context, logical_source_port)
            if logical_destination_port is not None:
                self._get_port(context, logical_destination_port)
            query = self._model_query(context, FlowClassifier)
            for flow_classifier_db in query.all():
                if self.flowclassifier_conflict(
                    fc,
                    flow_classifier_db
                ):
                    raise fc_ext.FlowClassifierInConflict(
                        id=flow_classifier_db['id']
                    )
            flow_classifier_db = FlowClassifier(
                id=uuidutils.generate_uuid(),
                project_id=project_id,
                name=fc['name'],
                description=fc['description'],
                ethertype=ethertype,
                protocol=protocol,
                source_port_range_min=source_port_range_min,
                source_port_range_max=source_port_range_max,
                destination_port_range_min=destination_port_range_min,
                destination_port_range_max=destination_port_range_max,
                source_ip_prefix=source_ip_prefix,
                destination_ip_prefix=destination_ip_prefix,
                logical_source_port=logical_source_port,
                logical_destination_port=logical_destination_port,
                l7_parameters=l7_parameters
            )
            context.session.add(flow_classifier_db)
            return self._make_flow_classifier_dict(flow_classifier_db)

    def _make_flow_classifier_dict(self, flow_classifier, fields=None):
        res = {
            'id': flow_classifier['id'],
            'name': flow_classifier['name'],
            'description': flow_classifier['description'],
            'project_id': flow_classifier['project_id'],
            'ethertype': flow_classifier['ethertype'],
            'protocol': flow_classifier['protocol'],
            'source_port_range_min': flow_classifier['source_port_range_min'],
            'source_port_range_max': flow_classifier['source_port_range_max'],
            'destination_port_range_min': (
                flow_classifier['destination_port_range_min']),
            'destination_port_range_max': (
                flow_classifier['destination_port_range_max']),
            'source_ip_prefix': flow_classifier['source_ip_prefix'],
            'destination_ip_prefix': flow_classifier[
                'destination_ip_prefix'],
            'logical_source_port': flow_classifier['logical_source_port'],
            'logical_destination_port': flow_classifier[
                'logical_destination_port'],
            'l7_parameters': {
                param['keyword']: param['value']
                for k, param in flow_classifier.l7_parameters.items()
            }
        }
        return self._fields(res, fields)

    @log_helpers.log_method_call
    def get_flow_classifiers(self, context, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
        marker_obj = self._get_marker_obj(context, 'flow_classifier',
                                          limit, marker)
        return self._get_collection(context,
                                    FlowClassifier,
                                    self._make_flow_classifier_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts,
                                    limit=limit, marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    @log_helpers.log_method_call
    def get_flow_classifier(self, context, id, fields=None):
        flow_classifier = self._get_flow_classifier(context, id)
        return self._make_flow_classifier_dict(flow_classifier, fields)

    def _get_flow_classifier(self, context, id):
        try:
            return self._get_by_id(context, FlowClassifier, id)
        except exc.NoResultFound:
            raise fc_ext.FlowClassifierNotFound(id=id)

    def _get_port(self, context, id):
        try:
            return self._get_by_id(context, models_v2.Port, id)
        except exc.NoResultFound:
            raise fc_ext.FlowClassifierPortNotFound(id=id)

    @log_helpers.log_method_call
    def update_flow_classifier(self, context, id, flow_classifier):
        new_fc = flow_classifier['flow_classifier']
        with db_api.context_manager.writer.using(context):
            old_fc = self._get_flow_classifier(context, id)
            old_fc.update(new_fc)
            return self._make_flow_classifier_dict(old_fc)

    @log_helpers.log_method_call
    def delete_flow_classifier(self, context, id):
        try:
            with db_api.context_manager.writer.using(context):
                fc = self._get_flow_classifier(context, id)
                context.session.delete(fc)
        except AssertionError:
            raise fc_ext.FlowClassifierInUse(id=id)
        except fc_ext.FlowClassifierNotFound:
            LOG.info("Deleting a non-existing flow classifier.")



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as n_const


STATUS_BUILDING = 'building'
STATUS_ACTIVE = 'active'
STATUS_ERROR = 'error'

SRC_NODE = 'src_node'
DST_NODE = 'dst_node'
SF_NODE = 'sf_node'

INSERTION_TYPE_L2 = 'l2'
INSERTION_TYPE_L3 = 'l3'
INSERTION_TYPE_BITW = 'bitw'
INSERTION_TYPE_TAP = 'tap'

MAX_HASH = 16

INSERTION_TYPE_DICT = {
    n_const.DEVICE_OWNER_ROUTER_HA_INTF: INSERTION_TYPE_L3,
    n_const.DEVICE_OWNER_ROUTER_INTF: INSERTION_TYPE_L3,
    n_const.DEVICE_OWNER_ROUTER_GW: INSERTION_TYPE_L3,
    n_const.DEVICE_OWNER_FLOATINGIP: INSERTION_TYPE_L3,
    n_const.DEVICE_OWNER_DHCP: INSERTION_TYPE_TAP,
    n_const.DEVICE_OWNER_DVR_INTERFACE: INSERTION_TYPE_L3,
    n_const.DEVICE_OWNER_AGENT_GW: INSERTION_TYPE_L3,
    n_const.DEVICE_OWNER_ROUTER_SNAT: INSERTION_TYPE_TAP,
    n_const.DEVICE_OWNER_LOADBALANCER: INSERTION_TYPE_TAP,
    'compute': INSERTION_TYPE_L2
}

ETH_TYPE_IP = 0x0800
ETH_TYPE_MPLS = 0x8847
ETH_TYPE_NSH = 0x894f



# Copyright (c) 2015 Huawei Technologies India Pvt.Limited.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutronclient.common import extension
from neutronclient.common import utils
from neutronclient.neutron import v2_0 as neutronv20

from networking_sfc._i18n import _

PORT_RESOURCE = 'port'
PORT_PAIR_RESOURCE = 'port_pair'


def get_port_id(client, id_or_name):
    return neutronv20.find_resourceid_by_name_or_id(client,
                                                    PORT_RESOURCE,
                                                    id_or_name)


def get_port_pair_id(client, id_or_name):
    return neutronv20.find_resourceid_by_name_or_id(client,
                                                    PORT_PAIR_RESOURCE,
                                                    id_or_name)


class PortPair(extension.NeutronClientExtension):
    resource = PORT_PAIR_RESOURCE
    resource_plural = '%ss' % resource
    object_path = '/sfc/%s' % resource_plural
    resource_path = '/sfc/%s/%%s' % resource_plural
    versions = ['2.0']


class PortPairCreate(extension.ClientExtensionCreate, PortPair):
    """Create a Port Pair."""

    shell_command = 'port-pair-create'

    def add_known_arguments(self, parser):
        parser.add_argument(
            'name',
            metavar='NAME',
            help=_('Name of the Port Pair.'))
        parser.add_argument(
            '--description',
            help=_('Description for the Port Pair.'))
        parser.add_argument(
            '--ingress',
            required=True,
            help=_('ID or name of the ingress neutron port.'))
        parser.add_argument(
            '--egress',
            required=True,
            help=_('ID or name of the egress neutron port.'))
        parser.add_argument(
            '--service-function-parameters',
            metavar='[correlation=CORRELATION_TYPE, weight=WEIGHT]',
            type=utils.str2dict_type(optional_keys=['correlation',
                                                    'weight']),
            help=_('Dictionary of Service function parameters. '
                   'Currently, only correlation=None|mpls and weight '
                   'is supported. Default correlation is None. Weight is '
                   'an integer that influences the selection'
                   'of a port pair within a port pair group '
                   'for a flow. The higher the weight, the more flows will '
                   'hash to the port pair. The default weight is 1.'))

    def args2body(self, parsed_args):
        body = {}
        client = self.get_client()
        if parsed_args.ingress:
            body['ingress'] = get_port_id(client, parsed_args.ingress)
        if parsed_args.egress:
            body['egress'] = get_port_id(client, parsed_args.egress)
        neutronv20.update_dict(parsed_args, body,
                               ['name', 'description',
                                'service_function_parameters'])
        return {self.resource: body}


class PortPairUpdate(extension.ClientExtensionUpdate, PortPair):
    """Update Port Pair's information."""

    shell_command = 'port-pair-update'

    def add_known_arguments(self, parser):
        parser.add_argument(
            '--name',
            metavar='NAME',
            help=_('Name of the Port Pair.'))
        parser.add_argument(
            '--description',
            help=_('Description for the Port Pair.'))

    def args2body(self, parsed_args):
        body = {}
        neutronv20.update_dict(parsed_args, body, ['name', 'description'])
        return {self.resource: body}


class PortPairDelete(extension.ClientExtensionDelete, PortPair):
    """Delete a given Port Pair."""

    shell_command = 'port-pair-delete'


class PortPairList(extension.ClientExtensionList, PortPair):
    """List Port Pairs that belongs to a given tenant."""

    shell_command = 'port-pair-list'
    list_columns = ['id', 'name', 'ingress', 'egress']
    pagination_support = True
    sorting_support = True


class PortPairShow(extension.ClientExtensionShow, PortPair):
    """Show information of a given Port Pair."""

    shell_command = 'port-pair-show'



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""modify_value_column_size_in_port_pair_group_params

Revision ID: 8329e9be2d8a
Revises: 61832141fb82
Create Date: 2017-04-19 15:13:29.833652

"""

# revision identifiers, used by Alembic.
revision = '8329e9be2d8a'
down_revision = '61832141fb82'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('sfc_port_pair_group_params', 'value',
                    existing_type=sa.String(255), type_=sa.String(1024))



# Copyright 2017 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from abc import ABCMeta
from abc import abstractmethod

import six

from neutron_lib.api import converters as lib_converters
from neutron_lib.api import extensions
from neutron_lib.api import validators as lib_validators
from neutron_lib.db import constants as db_const
from neutron_lib import exceptions as neutron_exc
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron.api import extensions as neutron_ext
from neutron.api.v2 import resource_helper

from networking_sfc._i18n import _
from networking_sfc import extensions as sfc_extensions
from networking_sfc.extensions import sfc as ext_sfc


cfg.CONF.import_opt('api_extensions_path', 'neutron.common.config')
neutron_ext.append_api_extensions_path(sfc_extensions.__path__)

SG_EXT = "service_graph"
SG_PREFIX = ext_sfc.SFC_PREFIX

SERVICE_GRAPH = 'service_graph'
SERVICE_GRAPHS = '%ss' % SERVICE_GRAPH


# NOTE(scsnow): move to neutron-lib
def validate_list_of_allowed_values(data, allowed_values=None):
    if not isinstance(data, list):
        msg = _("'%s' is not a list") % data
        return msg

    illegal_values = set(data) - set(allowed_values)
    if illegal_values:
        msg = _("Illegal values in a list: %s") % ', '.join(illegal_values)
        return msg


lib_validators.validators['type:list_of_allowed_values'] = \
    validate_list_of_allowed_values


class InvalidUUID(neutron_exc.InvalidInput):
    message = _(
        "An invalid UUID was specified: %%(error_message)s. "
        "Make sure only valid UUIDs are provided.")


class ServiceGraphInvalidPortChains(neutron_exc.InUse):
    message = _("Some of the Port Chain(s): %(port_chains)s, "
                "are already in use by a Service Graph.")


class ServiceGraphPortChainInUse(neutron_exc.InUse):
    message = _("Port Chain %(id)s in use.")


class ServiceGraphNotFound(neutron_exc.NotFound):
    message = _("Service Graph %(id)s not found.")


class ServiceGraphLoopDetected(neutron_exc.InvalidInput):
    message = _("Service Graph defined contains at least one port chain loop.")


class ServiceGraphInconsistentEncapsulation(neutron_exc.InvalidInput):
    message = _("Service Graph may only connect port-chains "
                "sharing the same correlation.")


class ServiceGraphImpossibleBranching(neutron_exc.InvalidInput):
    message = _("Service Graphs require source (branching) and destination "
                "port pair groups (their PPs) to have correlation enabled.")


class ServiceGraphFlowClassifierInConflict(neutron_exc.InvalidInput):
    message = _("Flow Classifier %(fc1_id)s conflicts with Flow Classifier "
                "%(fc2_id)s on one of the branching points being created.")


class ServiceGraphPortChainInConflict(neutron_exc.InvalidInput):
    message = _("Port Chain %(pc_id)s is duplicated on one "
                "of the branching points being created.")


def normalize_service_graph(port_chains):
    port_chains = lib_converters.convert_none_to_empty_dict(port_chains)
    for key in port_chains:
        if uuidutils.is_uuid_like(key):
            for val in port_chains[key]:
                if not uuidutils.is_uuid_like(val):
                    raise InvalidUUID(
                        error_message='UUID of destination Port-Chain '
                                      'is invalid: %s.' % key)
        else:
            raise InvalidUUID(
                error_message='UUID of source Port-Chain'
                              'is invalid: %s.' % key)
    return port_chains

RESOURCE_ATTRIBUTE_MAP = {
    SERVICE_GRAPHS: {
        'id': {
            'allow_post': False, 'allow_put': False,
            'is_visible': True,
            'validate': {'type:uuid': None},
            'primary_key': True},
        'name': {
            'allow_post': True, 'allow_put': True,
            'is_visible': True, 'default': '',
            'validate': {'type:string': db_const.NAME_FIELD_SIZE}},
        'description': {
            'allow_post': True, 'allow_put': True,
            'is_visible': True, 'default': '',
            'validate': {'type:string': db_const.DESCRIPTION_FIELD_SIZE}},
        'project_id': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True,
            'validate': {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
            'required_by_policy': True},
        'port_chains': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True,
            'validate': {'type:dict': None},
            'convert_to': normalize_service_graph}
    }
}

service_graph_quota_opts = [
    cfg.IntOpt('quota_service_graphs',
               default=10,
               help=_('maximum number of Service Graphs per project. '
                      'a negative value means unlimited.'))
]

cfg.CONF.register_opts(service_graph_quota_opts, 'QUOTAS')


class Servicegraph(extensions.ExtensionDescriptor):
    """Service Graph extension."""

    @classmethod
    def get_name(cls):
        return "Service Graph"

    @classmethod
    def get_alias(cls):
        return SG_EXT

    @classmethod
    def get_description(cls):
        return "Service Graph extension."

    @classmethod
    def get_updated(cls):
        return "2017-09-20T00:00:00-00:00"

    def update_attributes_map(self, attributes):
        super(Servicegraph, self).update_attributes_map(
            attributes, extension_attrs_map=RESOURCE_ATTRIBUTE_MAP)

    @classmethod
    def get_resources(cls):
        plural_mappings = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        plural_mappings['service_graphs'] = 'service_graph'
        return resource_helper.build_resource_info(
            plural_mappings,
            RESOURCE_ATTRIBUTE_MAP,
            ext_sfc.SFC_EXT,
            register_quota=True)

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        return {}


@six.add_metaclass(ABCMeta)
class ServiceGraphPluginBase(object):

    def get_plugin_name(self):
        return SG_EXT

    def get_plugin_type(self):
        return SG_EXT

    def get_plugin_description(self):
        return 'SFC Service Graphs extension for networking-sfc.'

    @abstractmethod
    def create_service_graph(self, context, service_graph):
        pass

    @abstractmethod
    def update_service_graph(self, context, id, service_graph):
        pass

    @abstractmethod
    def delete_service_graph(self, context, id):
        pass

    @abstractmethod
    def get_service_graphs(self, context, filters=None,
                           fields=None, sorts=None, limit=None,
                           marker=None, page_reverse=False):
        pass

    @abstractmethod
    def get_service_graph(self, context, id, fields=None):
        pass



# Copyright (c) 2017 One Convergence Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""extra attributes for pathnode

Revision ID: a3ad63aa834f
Revises: 8329e9be2d8a
Create Date: 2017-08-03 13:57:59.908621

"""

# revision identifiers, used by Alembic.
revision = 'a3ad63aa834f'
down_revision = 'd6fb381b65f2'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('sfc_path_nodes',
                  sa.Column('tap_enabled', sa.Boolean(), nullable=False,
                            server_default=sa.sql.false()))
    op.add_column('sfc_path_nodes',
                  sa.Column('previous_node_id', sa.String(length=36)))
    op.create_foreign_key('node_fk', 'sfc_path_nodes', 'sfc_path_nodes',
                          ['previous_node_id'], ['id'], ondelete='SET NULL')



# Copyright 2016 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import os.path
import signal


from oslo_config import cfg

from neutron.common import utils
from neutron.tests.functional import test_server

from networking_sfc.services.flowclassifier.common import (  # noqa
    config as fc_config)  # noqa
from networking_sfc.services.sfc.common import (  # noqa
    config as sfc_config)  # noqa


class TestService(test_server.TestPluginWorker):
    def _fake_start(self):
        with open(self.temp_file, 'ab') as f:
            f.write(test_server.FAKE_START_MSG)

    def _test_restart_service_on_sighup(self, service, workers=1):
        self._start_server(callback=service, workers=workers)
        os.kill(self.service_pid, signal.SIGHUP)
        expected_msg = test_server.FAKE_START_MSG * workers * 2
        expected_size = len(expected_msg)
        condition = lambda: (os.path.isfile(self.temp_file)
                             and os.stat(self.temp_file).st_size ==
                             expected_size)
        utils.wait_until_true(
            condition, timeout=5, sleep=0.1,
            exception=RuntimeError(
                "Timed out waiting for file %(filename)s to be created and "
                "its size become equal to %(size)s." %
                {'filename': self.temp_file,
                 'size': expected_size}))
        with open(self.temp_file, 'rb') as f:
            res = f.readline()
            self.assertEqual(expected_msg, res)

    def test_start(self):
        cfg.CONF.set_override(
            'service_plugins', [
                'networking_sfc.services.flowclassifier.plugin.'
                'FlowClassifierPlugin',
                'networking_sfc.services.sfc.plugin.SfcPlugin'
            ]
        )
        cfg.CONF.set_override(
            'drivers', ['ovs'], group='flowclassifier'
        )
        cfg.CONF.set_override(
            'drivers', ['ovs'], group='sfc'
        )
        super(TestService, self).test_start()



# Copyright 2015 Futurewei.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Exceptions used by FlowClassifier plugin and drivers."""

from neutron_lib import exceptions

from networking_sfc._i18n import _


class FlowClassifierDriverError(exceptions.NeutronException):
    """flow classifier driver call failed."""
    message = _("%(method)s failed.")


class FlowClassifierException(exceptions.NeutronException):
    """Base for flow classifier driver exceptions returned to user."""
    pass


class FlowClassifierBadRequest(exceptions.BadRequest, FlowClassifierException):
    """Base for flow classifier driver bad request exceptions."""
    message = _("%(message)s")



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log_helpers

from networking_sfc.services.flowclassifier.drivers import base as fc_driver


class DummyDriver(fc_driver.FlowClassifierDriverBase):
    """Flow Classifier Driver Dummy Class."""
    def initialize(self):
        pass

    @log_helpers.log_method_call
    def create_flow_classifier(self, context):
        pass

    @log_helpers.log_method_call
    def update_flow_classifier(self, context):
        pass

    @log_helpers.log_method_call
    def delete_flow_classifier(self, context):
        pass

    @log_helpers.log_method_call
    def create_flow_classifier_precommit(self, context):
        pass
